@@ -33,8 +33,10 @@ from fido2.ctap2 import (
     CredentialManagement,
     FPBioEnrollment,
     CaptureError,
+    Config,
 )
 from fido2.pcsc import CtapPcscDevice
+from yubikit.management import CAPABILITY
 from yubikit.core.fido import FidoConnection
 from yubikit.core.smartcard import SW
 from time import sleep
@@ -45,13 +47,14 @@ from .util import (
     click_group,
     prompt_timeout,
     is_yk4_fips,
+    pretty_print,
 )
 from .util import CliFail
 from ..fido import is_in_fips_mode, fips_reset, fips_change_pin, fips_verify_pin
 from ..hid import list_ctap_devices
 from ..pcsc import list_devices as list_ccid
 from smartcard.Exceptions import NoCardException, CardConnectionException
-from typing import Optional, Sequence, List
+from typing import Optional, Sequence, List, Dict
 
 import io
 import csv as _csv
@@ -61,10 +64,6 @@ import logging
 logger = logging.getLogger(__name__)
 
 
-FIPS_PIN_MIN_LENGTH = 6
-PIN_MIN_LENGTH = 4
-
-
 @click_group(connections=[FidoConnection])
 @click.pass_context
 @click_postpone_execution
@@ -101,54 +100,59 @@ def info(ctx):
     """
     conn = ctx.obj["conn"]
     ctap2 = ctx.obj.get("ctap2")
+    info: Dict = {}
+    lines: List = [info]
 
     if is_yk4_fips(ctx.obj["info"]):
-        click.echo("FIPS Approved Mode: " + ("Yes" if is_in_fips_mode(conn) else "No"))
+        info["FIPS Approved Mode"] = "Yes" if is_in_fips_mode(conn) else "No"
     elif ctap2:
         client_pin = ClientPin(ctap2)  # N.B. All YubiKeys with CTAP2 support PIN.
         if ctap2.info.options["clientPin"]:
             if ctap2.info.force_pin_change:
-                click.echo(
+                lines.append(
                     "NOTE: The FIDO PIN is disabled and must be changed before it can "
                     "be used!"
                 )
             pin_retries, power_cycle = client_pin.get_pin_retries()
             if pin_retries:
-                click.echo(f"PIN is set, with {pin_retries} attempt(s) remaining.")
+                info["PIN"] = f"{pin_retries} attempt(s) remaining"
                 if power_cycle:
-                    click.echo(
+                    lines.append(
                         "PIN is temporarily blocked. "
                         "Remove and re-insert the YubiKey to unblock."
                     )
             else:
-                click.echo("PIN is set, but has been blocked.")
+                info["PIN"] = "blocked"
         else:
-            click.echo("PIN is not set.")
+            info["PIN"] = "not set"
+        info["Minimum PIN length"] = ctap2.info.min_pin_length
 
         bio_enroll = ctap2.info.options.get("bioEnroll")
         if bio_enroll:
             uv_retries = client_pin.get_uv_retries()
             if uv_retries:
-                click.echo(
-                    f"Fingerprints registered, with {uv_retries} attempt(s) "
-                    "remaining."
-                )
+                info["Fingerprints"] = f"registered, {uv_retries} attempt(s) remaining"
             else:
-                click.echo(
-                    "Fingerprints registered, but blocked until PIN is verified."
-                )
+                info["Fingerprints"] = "registered, blocked until PIN is verified"
         elif bio_enroll is False:
-            click.echo("No fingerprints have been registered.")
+            info["Fingerprints"] = "not registered"
 
         always_uv = ctap2.info.options.get("alwaysUv")
         if always_uv is not None:
-            click.echo(
-                "Always Require User Verification is turned "
-                + ("on." if always_uv else "off.")
-            )
+            info["Always Require UV"] = "on" if always_uv else "off"
+
+        remaining_creds = ctap2.info.remaining_disc_creds
+        if remaining_creds is not None:
+            info["Credential storage remaining"] = remaining_creds
+
+        ep = ctap2.info.options.get("ep")
+        if ep is not None:
+            info["Enterprise Attestation"] = "enabled" if ep else "disabled"
 
     else:
-        click.echo("PIN is not supported.")
+        info["PIN"] = "not supported"
+
+    click.echo("\n".join(pretty_print(lines)))
 
 
 @fido.command("reset")
@@ -165,8 +169,14 @@ def reset(ctx, force):
     inserted, and requires a touch on the YubiKey.
     """
 
-    conn = ctx.obj["conn"]
+    info = ctx.obj["info"]
+    if CAPABILITY.FIDO2 in info.reset_blocked:
+        raise CliFail(
+            "Cannot perform FIDO reset when PIV is configured, "
+            "use 'ykman config reset' for full factory reset."
+        )
 
+    conn = ctx.obj["conn"]
     if isinstance(conn, CtapPcscDevice):  # NFC
         readers = list_ccid(conn._name)
         if not readers or readers[0].reader.name != conn._name:
@@ -226,10 +236,10 @@ def reset(ctx, force):
         )
         if is_fips:
             destroy_input = click_prompt(
-                "WARNING! This is a YubiKey FIPS device. This command will also "
-                "overwrite the U2F attestation key; this action cannot be undone and "
-                "this YubiKey will no longer be a FIPS compliant device.\n"
-                'To proceed, please enter the text "OVERWRITE"',
+                "WARNING! This is a YubiKey FIPS (4 Series) device. This command will "
+                "also overwrite the U2F attestation key; this action cannot be undone "
+                "and this YubiKey will no longer be a FIPS compliant device.\n"
+                'To proceed, enter the text "OVERWRITE"',
                 default="",
                 show_default=False,
             )
@@ -298,16 +308,17 @@ def access():
     "-u",
     "--u2f",
     is_flag=True,
-    help="set FIDO U2F PIN instead of FIDO2 PIN (YubiKey 4 FIPS only)",
+    help="set FIDO U2F PIN instead of FIDO2 PIN (YubiKey FIPS only)",
 )
 def change_pin(ctx, pin, new_pin, u2f):
     """
     Set or change the PIN code.
 
-    The FIDO2 PIN must be at least 4 characters long, and supports any type
-    of alphanumeric characters.
+    The FIDO2 PIN must be at least 4 characters long, and supports any type of
+    alphanumeric characters. Some YubiKeys can be configured to require a longer
+    PIN.
 
-    On YubiKey FIPS, a PIN can be set for FIDO U2F. That PIN must be at least
+    On YubiKey FIPS (4 Series), a PIN can be set for FIDO U2F. That PIN must be at least
     6 characters long.
     """
 
@@ -315,22 +326,29 @@ def change_pin(ctx, pin, new_pin, u2f):
 
     if is_fips and not u2f:
         raise CliFail(
-            "This is a YubiKey FIPS. To set the U2F PIN, pass the --u2f option."
+            "This is a YubiKey FIPS (4 Series). "
+            "To set the U2F PIN, pass the --u2f option."
         )
 
     if u2f and not is_fips:
         raise CliFail(
-            "This is not a YubiKey 4 FIPS, and therefore does not support a U2F PIN. "
-            "To set the FIDO2 PIN, remove the --u2f option."
+            "This is not a YubiKey FIPS (4 Series), and therefore does not support a "
+            "U2F PIN. To set the FIDO2 PIN, remove the --u2f option."
         )
 
     if is_fips:
         conn = ctx.obj["conn"]
+        min_len = 6
     else:
         ctap2 = ctx.obj.get("ctap2")
         if not ctap2:
             raise CliFail("PIN is not supported on this YubiKey.")
         client_pin = ClientPin(ctap2)
+        min_len = ctap2.info.min_pin_length
+
+    def _fail_if_not_valid_pin(pin=None, name="PIN"):
+        if not pin or len(pin) < min_len:
+            raise CliFail(f"{name} must be at least {min_len} characters long")
 
     def prompt_new_pin():
         return click_prompt(
@@ -340,8 +358,6 @@ def change_pin(ctx, pin, new_pin, u2f):
         )
 
     def change_pin(pin, new_pin):
-        if pin is not None:
-            _fail_if_not_valid_pin(ctx, pin, is_fips)
         try:
             if is_fips:
                 try:
@@ -350,7 +366,7 @@ def change_pin(ctx, pin, new_pin, u2f):
                 except ApduError as e:
                     if e.code == SW.WRONG_LENGTH:
                         pin = _prompt_current_pin()
-                        _fail_if_not_valid_pin(ctx, pin, is_fips)
+                        _fail_if_not_valid_pin(pin)
                         fips_change_pin(conn, pin, new_pin)
                     else:
                         raise
@@ -360,7 +376,7 @@ def change_pin(ctx, pin, new_pin, u2f):
 
         except CtapError as e:
             if e.code == CtapError.ERR.PIN_POLICY_VIOLATION:
-                raise CliFail("New PIN doesn't meet policy requirements.")
+                raise CliFail("New PIN doesn't meet complexity requirements.")
             else:
                 _fail_pin_error(ctx, e, "Failed to change PIN: %s")
 
@@ -373,12 +389,12 @@ def change_pin(ctx, pin, new_pin, u2f):
                 raise CliFail(f"Failed to change PIN: SW={e.code:04x}")
 
     def set_pin(new_pin):
-        _fail_if_not_valid_pin(ctx, new_pin, is_fips)
+        _fail_if_not_valid_pin(new_pin)
         try:
             client_pin.set_pin(new_pin)
         except CtapError as e:
             if e.code == CtapError.ERR.PIN_POLICY_VIOLATION:
-                raise CliFail("New PIN doesn't meet policy requirements.")
+                raise CliFail("New PIN doesn't meet complexity requirements.")
             else:
                 raise CliFail(f"Failed to set PIN: {e.code}")
 
@@ -392,14 +408,11 @@ def change_pin(ctx, pin, new_pin, u2f):
 
     if not new_pin:
         new_pin = prompt_new_pin()
+    _fail_if_not_valid_pin(new_pin, "New PIN")
 
     if is_fips:
-        _fail_if_not_valid_pin(ctx, new_pin, is_fips)
         change_pin(pin, new_pin)
     else:
-        min_len = ctap2.info.min_pin_length
-        if len(new_pin) < min_len:
-            raise CliFail("New PIN is too short. Minimum length: {min_len}")
         if ctap2.info.options.get("clientPin"):
             change_pin(pin, new_pin)
         else:
@@ -428,7 +441,7 @@ def verify(ctx, pin):
     Verify the FIDO PIN against a YubiKey.
 
     For YubiKeys supporting FIDO2 this will reset the "retries" counter of the PIN.
-    For YubiKey FIPS this will unlock the session, allowing U2F registration.
+    For YubiKey FIPS (4 Series) this will unlock the session, allowing U2F registration.
     """
 
     ctap2 = ctx.obj.get("ctap2")
@@ -443,7 +456,6 @@ def verify(ctx, pin):
         except CtapError as e:
             raise CliFail(f"PIN verification failed: {e}")
     elif is_yk4_fips(ctx.obj["info"]):
-        _fail_if_not_valid_pin(ctx, pin, True)
         try:
             fips_verify_pin(ctx.obj["conn"], pin)
         except ApduError as e:
@@ -460,16 +472,84 @@ def verify(ctx, pin):
     click.echo("PIN verified.")
 
 
+def _init_config(ctx, pin):
+    ctap2 = ctx.obj.get("ctap2")
+    if not Config.is_supported(ctap2.info):
+        raise CliFail("Authenticator Configuration is not supported on this YubiKey.")
+
+    protocol = None
+    token = None
+    if ctap2.info.options.get("clientPin"):
+        pin = _require_pin(ctx, pin, "Authenticator Configuration")
+        client_pin = ClientPin(ctap2)
+        try:
+            protocol = client_pin.protocol
+            token = client_pin.get_pin_token(
+                pin, ClientPin.PERMISSION.AUTHENTICATOR_CFG
+            )
+        except CtapError as e:
+            _fail_pin_error(ctx, e, "PIN error: %s")
+
+    return Config(ctap2, protocol, token)
+
+
+@access.command("force-change")
+@click.pass_context
+@click.option("-P", "--pin", help="PIN code")
+def force_pin_change(ctx, pin):
+    """
+    Force the PIN to be changed to a new value before use.
+    """
+    options = ctx.obj.get("ctap2").info.options
+    if not options.get("setMinPINLength"):
+        raise CliFail("Force change PIN is not supported on this YubiKey.")
+    if not options.get("clientPin"):
+        raise CliFail("No PIN is set.")
+
+    config = _init_config(ctx, pin)
+    config.set_min_pin_length(force_change_pin=True)
+
+
+@access.command("set-min-length")
+@click.pass_context
+@click.option("-P", "--pin", help="PIN code")
+@click.option("-R", "--rp-id", multiple=True, help="RP ID to allow")
+@click.argument("length", type=click.IntRange(4))
+def set_min_pin_length(ctx, pin, rp_id, length):
+    """
+    Set the minimum length allowed for PIN.
+
+    Optionally use the --rp option to specify which RPs are allowed to request this
+    information.
+    """
+    info = ctx.obj["ctap2"].info
+    if not info.options.get("setMinPINLength"):
+        raise CliFail("Set minimum PIN length is not supported on this YubiKey.")
+    if info.options.get("alwaysUv") and not info.options.get("clientPin"):
+        raise CliFail(
+            "Setting min PIN length requires a PIN to be set when alwaysUv is enabled."
+        )
+
+    min_len = info.min_pin_length
+    if length < min_len:
+        raise CliFail(f"Cannot set a minimum length that is shorter than {min_len}.")
+
+    config = _init_config(ctx, pin)
+    if rp_id:
+        ctap2 = ctx.obj.get("ctap2")
+        cap = ctap2.info.max_rpids_for_min_pin
+        if len(rp_id) > cap:
+            raise CliFail(
+                f"Authenticator supports up to {cap} RP IDs ({len(rp_id)} given)."
+            )
+
+    config.set_min_pin_length(min_pin_length=length, rp_ids=rp_id)
+
+
 def _prompt_current_pin(prompt="Enter your current PIN"):
     return click_prompt(prompt, hide_input=True)
 
 
-def _fail_if_not_valid_pin(ctx, pin=None, is_fips=False):
-    min_length = FIPS_PIN_MIN_LENGTH if is_fips else PIN_MIN_LENGTH
-    if not pin or len(pin) < min_length:
-        ctx.fail(f"PIN must be over {min_length} characters long")
-
-
 def _gen_creds(credman):
     data = credman.get_metadata()
     if data.get(CredentialManagement.RESULT.EXISTING_CRED_COUNT) == 0:
@@ -791,3 +871,45 @@ def bio_delete(ctx, template_id, pin, force):
             logger.info("Fingerprint template deleted")
         except CtapError as e:
             raise CliFail(f"Failed to delete fingerprint: {e.code.name}")
+
+
+@fido.group("config")
+def config():
+    """
+    Manage FIDO configuration.
+    """
+
+
+@config.command("toggle-always-uv")
+@click.pass_context
+@click.option("-P", "--pin", help="PIN code")
+def toggle_always_uv(ctx, pin):
+    """
+    Toggles the state of Always Require User Verification.
+    """
+    options = ctx.obj.get("ctap2").info.options
+    if "alwaysUv" not in options:
+        raise CliFail("Always Require UV is not supported on this YubiKey.")
+
+    config = _init_config(ctx, pin)
+    config.toggle_always_uv()
+
+
+@config.command("enable-ep-attestation")
+@click.pass_context
+@click.option("-P", "--pin", help="PIN code")
+def enable_ep_attestation(ctx, pin):
+    """
+    Enables Enterprise Attestation for Authenticators pre-configured to support it.
+    """
+    options = ctx.obj.get("ctap2").info.options
+    if "ep" not in options:
+        raise CliFail("Enterprise Attestation is not supported on this YubiKey.")
+    if options.get("alwaysUv") and not options.get("clientPin"):
+        raise CliFail(
+            "Enabling Enterprise Attestation requires a PIN to be set when alwaysUv is "
+            "enabled."
+        )
+
+    config = _init_config(ctx, pin)
+    config.enable_enterprise_attestation()
